﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;

  using Interpreter;
  using Stack;

  /// <summary>
  ///     Empties the given stack.
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class FlushExpression<T> : StatelessExpression {
    public bool Eval(IPushStack<T> stack) {
      stack.Clear();
      return true;
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.FLUSH")]
  public class IntegerFlushExpression : FlushExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.FLUSH")]
  public class FloatFlushExpression : FlushExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.FLUSH")]
  public class BooleanFlushExpression : FlushExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.FLUSH")]
  public class NameFlushExpression : FlushExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.NameStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.FLUSH")]
  public class ExecFlushExpression : FlushExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.ExecStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.FLUSH")]
  public class CodeFlushExpression : FlushExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CodeStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.FLUSH")]
  public class CharFlushExpression : FlushExpression<char> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CharStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.FLUSH")]
  public class StringFlushExpression : FlushExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].FLUSH")]
  public class IntegerVectorFlushExpression : FlushExpression<List<long>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerVectorStack);
    }
  }
}