﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;

  using Interpreter;
  using Stack;

  public abstract class ShoveExpression<T> : StatelessExpression {
    protected bool Eval(IPushStack<T> stack, IPushStack<long> integerStack, int count = 1) {
      if (integerStack.Count == 0 ||
          integerStack.Top > stack.Count - count ||
          integerStack.Top < 0 ||
          stack.Count < 2)
        return false;

      var index = (int)integerStack.Pop();
      var item = stack.Pop();

      stack.Insert(index, item);
      return true;
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.SHOVE")]
  public class IntegerShoveExpression : ShoveExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, interpreter.IntegerStack, 2);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.SHOVE", StackTypes.Integer)]
  public class FloatShoveExpression : ShoveExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.SHOVE", StackTypes.Integer)]
  public class BooleanShoveExpression : ShoveExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.SHOVE", StackTypes.Integer)]
  public class NameShoveExpression : ShoveExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.NameStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.SHOVE", StackTypes.Integer)]
  public class ExecShoveExpression : ShoveExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.ExecStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.SHOVE", StackTypes.Integer)]
  public class CodeShoveExpression : ShoveExpression<Expression> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CodeStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.SHOVE", StackTypes.Integer)]
  public class CharShoveExpression : ShoveExpression<char> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.CharStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.SHOVE", StackTypes.Integer)]
  public class StringShoveExpression : ShoveExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.StringStack, interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].SHOVE", StackTypes.Integer)]
  public class IntegerVectorShoveExpression : ShoveExpression<List<long>> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }
}