﻿using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Counts the occurrences of the top lit-type item in the top type vector.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  public abstract class VectorOccurrenceOfExpression<T> : StatelessExpression {
    protected bool Eval(
      IInternalPushInterpreter interpreter,
      IPushStack<List<T>> vectorStack,
      IPushStack<T> literalStack,
      bool canOverride = false) {
      if (vectorStack.IsEmpty || literalStack.IsEmpty)
        return false;

      var vector = vectorStack.Pop();
      var literal = canOverride ? literalStack.Top : literalStack.Pop();

      var occurrence = 0;
      for (var i = 0; i < vector.Count; i++) {
        if (vector[i].Equals(literal)) occurrence++;
      }

      if (canOverride) interpreter.IntegerStack.SetTop(occurrence);
      else interpreter.IntegerStack.Push(occurrence);
      return true;
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].OCCURRENCEOF", StackTypes.Integer)]
  public class IntegerVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<long> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack, true);
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].OCCURRENCEOF", StackTypes.Float | StackTypes.Integer)]
  public class FloatVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].OCCURRENCEOF", StackTypes.Boolean | StackTypes.Integer)]
  public class BooleanVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<bool> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].OCCURRENCEOF", StackTypes.String | StackTypes.Integer)]
  public class StringVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<string> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
